# SPDX-FileCopyrightText: 2010-2022 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later


def create_and_link_mesh(name, faces, face_nors, points, global_matrix):
    """
    Create a blender mesh and object called name from a list of
    *points* and *faces* and link it in the current scene.
    """

    import array
    from itertools import chain
    import bpy

    mesh = bpy.data.meshes.new(name)
    mesh.from_pydata(points, [], faces)

    if face_nors:
        # Note: we store 'temp' normals in loops, since validate() may alter final mesh,
        #       we can only set custom lnors *after* calling it.
        mesh.create_normals_split()
        lnors = tuple(chain(*chain(*zip(face_nors, face_nors, face_nors))))
        mesh.loops.foreach_set("normal", lnors)

    mesh.transform(global_matrix)

    # update mesh to allow proper display
    mesh.validate(clean_customdata=False)  # *Very* important to not remove lnors here!

    if face_nors:
        clnors = array.array('f', [0.0] * (len(mesh.loops) * 3))
        mesh.loops.foreach_get("normal", clnors)

        mesh.polygons.foreach_set("use_smooth", [True] * len(mesh.polygons))

        mesh.normals_split_custom_set(tuple(zip(*(iter(clnors),) * 3)))
        mesh.use_auto_smooth = True
        mesh.free_normals_split()

    mesh.update()

    obj = bpy.data.objects.new(name, mesh)
    bpy.context.collection.objects.link(obj)
    bpy.context.view_layer.objects.active = obj
    obj.select_set(True)


def faces_from_mesh(ob, global_matrix, use_mesh_modifiers=False):
    """
    From an object, return a generator over a list of faces.

    Each faces is a list of his vertices. Each vertex is a tuple of
    his coordinate.

    use_mesh_modifiers
        Apply the preview modifier to the returned liste

    triangulate
        Split the quad into two triangles
    """

    import bpy

    # get the editmode data
    if ob.mode == "EDIT":
        ob.update_from_editmode()

    # get the modifiers
    if use_mesh_modifiers:
        depsgraph = bpy.context.evaluated_depsgraph_get()
        mesh_owner = ob.evaluated_get(depsgraph)
    else:
        mesh_owner = ob

    # Object.to_mesh() is not guaranteed to return a mesh.
    try:
        mesh = mesh_owner.to_mesh()
    except RuntimeError:
        return

    if mesh is None:
        return

    mat = global_matrix @ ob.matrix_world
    mesh.transform(mat)
    if mat.is_negative:
        mesh.flip_normals()
    mesh.calc_loop_triangles()

    vertices = mesh.vertices

    for tri in mesh.loop_triangles:
        yield [vertices[index].co.copy() for index in tri.vertices]

    mesh_owner.to_mesh_clear()



# SPDX-FileCopyrightText: 2019-2023 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Passepartout on Specials menu

The passepartout value of local cameras is now available on the Specials
menu for easy access.
Under Specials menu W, when in Camera view.
"""

import bpy


def button_camera_passepartout(self, context):
    view3d = context.space_data.region_3d
    cam = context.scene.camera

    if view3d.view_perspective == "CAMERA":
        if cam is None or not hasattr(cam, "data") or cam.type != "CAMERA":
            return

        layout = self.layout
        if cam.data.show_passepartout:
            layout.prop(cam.data, "passepartout_alpha", text="Passepartout")
        else:
            layout.prop(cam.data, "show_passepartout")


def register():
    bpy.types.VIEW3D_MT_object_context_menu.append(button_camera_passepartout)


def unregister():
    bpy.types.VIEW3D_MT_object_context_menu.remove(button_camera_passepartout)



# SPDX-FileCopyrightText: 2023 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later

import bpy
from bpy_extras.node_utils import connect_sockets
from math import hypot


def force_update(context):
    context.space_data.node_tree.update_tag()


def dpi_fac():
    prefs = bpy.context.preferences.system
    return prefs.dpi / 72


def prefs_line_width():
    prefs = bpy.context.preferences.system
    return prefs.pixel_size


def node_mid_pt(node, axis):
    if axis == 'x':
        d = node.location.x + (node.dimensions.x / 2)
    elif axis == 'y':
        d = node.location.y - (node.dimensions.y / 2)
    else:
        d = 0
    return d


def autolink(node1, node2, links):
    available_inputs = [inp for inp in node2.inputs if inp.enabled]
    available_outputs = [outp for outp in node1.outputs if outp.enabled]
    for outp in available_outputs:
        for inp in available_inputs:
            if not inp.is_linked and inp.name == outp.name:
                connect_sockets(outp, inp)
                return True

    for outp in available_outputs:
        for inp in available_inputs:
            if not inp.is_linked and inp.type == outp.type:
                connect_sockets(outp, inp)
                return True

    # force some connection even if the type doesn't match
    if available_outputs:
        for inp in available_inputs:
            if not inp.is_linked:
                connect_sockets(available_outputs[0], inp)
                return True

    # even if no sockets are open, force one of matching type
    for outp in available_outputs:
        for inp in available_inputs:
            if inp.type == outp.type:
                connect_sockets(outp, inp)
                return True

    # do something!
    for outp in available_outputs:
        for inp in available_inputs:
            connect_sockets(outp, inp)
            return True

    print("Could not make a link from " + node1.name + " to " + node2.name)
    return False


def abs_node_location(node):
    abs_location = node.location
    if node.parent is None:
        return abs_location
    return abs_location + abs_node_location(node.parent)


def node_at_pos(nodes, context, event):
    nodes_under_mouse = []
    target_node = None

    store_mouse_cursor(context, event)
    x, y = context.space_data.cursor_location

    # Make a list of each corner (and middle of border) for each node.
    # Will be sorted to find nearest point and thus nearest node
    node_points_with_dist = []
    for node in nodes:
        skipnode = False
        if node.type != 'FRAME':  # no point trying to link to a frame node
            dimx = node.dimensions.x / dpi_fac()
            dimy = node.dimensions.y / dpi_fac()
            locx, locy = abs_node_location(node)

            if not skipnode:
                node_points_with_dist.append([node, hypot(x - locx, y - locy)])  # Top Left
                node_points_with_dist.append([node, hypot(x - (locx + dimx), y - locy)])  # Top Right
                node_points_with_dist.append([node, hypot(x - locx, y - (locy - dimy))])  # Bottom Left
                node_points_with_dist.append([node, hypot(x - (locx + dimx), y - (locy - dimy))])  # Bottom Right

                node_points_with_dist.append([node, hypot(x - (locx + (dimx / 2)), y - locy)])  # Mid Top
                node_points_with_dist.append([node, hypot(x - (locx + (dimx / 2)), y - (locy - dimy))])  # Mid Bottom
                node_points_with_dist.append([node, hypot(x - locx, y - (locy - (dimy / 2)))])  # Mid Left
                node_points_with_dist.append([node, hypot(x - (locx + dimx), y - (locy - (dimy / 2)))])  # Mid Right

    nearest_node = sorted(node_points_with_dist, key=lambda k: k[1])[0][0]

    for node in nodes:
        if node.type != 'FRAME' and skipnode == False:
            locx, locy = abs_node_location(node)
            dimx = node.dimensions.x / dpi_fac()
            dimy = node.dimensions.y / dpi_fac()
            if (locx <= x <= locx + dimx) and \
               (locy - dimy <= y <= locy):
                nodes_under_mouse.append(node)

    if len(nodes_under_mouse) == 1:
        if nodes_under_mouse[0] != nearest_node:
            target_node = nodes_under_mouse[0]  # use the node under the mouse if there is one and only one
        else:
            target_node = nearest_node  # else use the nearest node
    else:
        target_node = nearest_node
    return target_node


def store_mouse_cursor(context, event):
    space = context.space_data
    v2d = context.region.view2d
    tree = space.edit_tree

    # convert mouse position to the View2D for later node placement
    if context.region.type == 'WINDOW':
        space.cursor_location_from_region(event.mouse_region_x, event.mouse_region_y)
    else:
        space.cursor_location = tree.view_center


def get_active_tree(context):
    tree = context.space_data.node_tree
    path = []
    # Get nodes from currently edited tree.
    # If user is editing a group, space_data.node_tree is still the base level (outside group).
    # context.active_node is in the group though, so if space_data.node_tree.nodes.active is not
    # the same as context.active_node, the user is in a group.
    # Check recursively until we find the real active node_tree:
    if tree.nodes.active:
        while tree.nodes.active != context.active_node:
            tree = tree.nodes.active.node_tree
            path.append(tree)
    return tree, path


def get_nodes_links(context):
    tree, path = get_active_tree(context)
    return tree.nodes, tree.links


viewer_socket_name = "tmp_viewer"


def is_viewer_socket(socket):
    # checks if a internal socket is a valid viewer socket
    return socket.name == viewer_socket_name and socket.NWViewerSocket


def get_internal_socket(socket):
    # get the internal socket from a socket inside or outside the group
    node = socket.node
    if node.type == 'GROUP_OUTPUT':
        iterator = node.id_data.interface.items_tree
    elif node.type == 'GROUP_INPUT':
        iterator = node.id_data.interface.items_tree
    elif hasattr(node, "node_tree"):
        iterator = node.node_tree.interface.items_tree
    else:
        return None

    for s in iterator:
        if s.identifier == socket.identifier:
            return s
    return iterator[0]


def is_viewer_link(link, output_node):
    if link.to_node == output_node and link.to_socket == output_node.inputs[0]:
        return True
    if link.to_node.type == 'GROUP_OUTPUT':
        socket = get_internal_socket(link.to_socket)
        if is_viewer_socket(socket):
            return True
    return False


def get_group_output_node(tree):
    for node in tree.nodes:
        if node.type == 'GROUP_OUTPUT' and node.is_active_output:
            return node


def get_output_location(tree):
    # get right-most location
    sorted_by_xloc = (sorted(tree.nodes, key=lambda x: x.location.x))
    max_xloc_node = sorted_by_xloc[-1]

    # get average y location
    sum_yloc = 0
    for node in tree.nodes:
        sum_yloc += node.location.y

    loc_x = max_xloc_node.location.x + max_xloc_node.dimensions.x + 80
    loc_y = sum_yloc / len(tree.nodes)
    return loc_x, loc_y


def nw_check(context):
    space = context.space_data
    valid_trees = ["ShaderNodeTree", "CompositorNodeTree", "TextureNodeTree", "GeometryNodeTree"]

    if (space.type == 'NODE_EDITOR'
            and space.node_tree is not None
            and space.node_tree.library is None
            and space.tree_type in valid_trees):
        return True

    return False


def get_first_enabled_output(node):
    for output in node.outputs:
        if output.enabled:
            return output
    else:
        return node.outputs[0]


def is_visible_socket(socket):
    return not socket.hide and socket.enabled and socket.type != 'CUSTOM'


class NWBase:
    @classmethod
    def poll(cls, context):
        return nw_check(context)



# SPDX-FileCopyrightText: 2010-2022 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later

__author__ = "Nutti <nutti.metro@gmail.com>"
__status__ = "production"
__version__ = "6.6"
__date__ = "22 Apr 2022"

if "bpy" in locals():
    import importlib
    importlib.reload(bl_class_registry)
    importlib.reload(compatibility)
    importlib.reload(graph)
    importlib.reload(property_class_registry)
else:
    from . import bl_class_registry
    from . import compatibility
    from . import graph
    from . import property_class_registry

import bpy



# SPDX-FileCopyrightText: 2011 Ryan Inch
#
# SPDX-License-Identifier: GPL-2.0-or-later

import time
from math import cos, sin, pi, floor
import bpy
import blf
import gpu
from gpu_extras.batch import batch_for_shader

from bpy.types import Operator

from . import internals

from .qcd_operators import (
    get_move_selection,
    get_move_active,
    )

def spacer():
    spacer = 10
    return round(spacer * scale_factor())

def scale_factor():
    return bpy.context.preferences.system.ui_scale

def get_coords(area):
    x = area["vert"][0]
    y = area["vert"][1]
    w = area["width"]
    h = area["height"]

    vertices = (
        (x, y-h), # bottom left
        (x+w, y-h), # bottom right
        (x, y), # top left
        (x+w, y)) # top right

    indices = (
        (0, 1, 2), (2, 1, 3))

    return vertices, indices

def get_x_coords(area):
    x = area["vert"][0]
    y = area["vert"][1]
    w = area["width"]
    h = area["height"]

    vertices = (
        (x, y), # top left A
        (x+(w*0.1), y), # top left B
        (x+w, y), # top right A
        (x+w-(w*0.1), y), # top right B
        (x, y-h), # bottom left A
        (x+(w*0.1), y-h), # bottom left B
        (x+w, y-h), # bottom right A
        (x+w-(w*0.1), y-h), # bottom right B
        (x+(w/2)-(w*0.05), y-(h/2)), # center left
        (x+(w/2)+(w*0.05), y-(h/2))  # center right
        )

    indices = (
        (0,1,8), (1,8,9), # top left bar
        (2,3,9), (3,9,8), # top right bar
        (4,5,8), (5,8,9), # bottom left bar
        (6,7,8), (6,9,8)  # bottom right bar
        )

    return vertices, indices

def get_circle_coords(area):
    # set x, y to center
    x = area["vert"][0] + area["width"] / 2
    y = area["vert"][1] - area["width"] / 2
    radius = area["width"] / 2
    sides = 32
    vertices = [(radius * cos(side * 2 * pi / sides) + x,
                 radius * sin(side * 2 * pi / sides) + y)
                 for side in range(sides + 1)]

    return vertices

def draw_rounded_rect(area, shader, color, tl=5, tr=5, bl=5, br=5, outline=False):
    sides = 32

    tl = round(tl * scale_factor())
    tr = round(tr * scale_factor())
    bl = round(bl * scale_factor())
    br = round(br * scale_factor())

    gpu.state.blend_set('ALPHA')

    if outline:
        thickness = round(2 * scale_factor())
        thickness = max(thickness, 2)
        shader.uniform_float("lineWidth", thickness)

    draw_type = 'TRI_FAN' if not outline else 'LINE_STRIP'

    # top left corner
    vert_x = area["vert"][0] + tl
    vert_y = area["vert"][1] - tl
    tl_vert = (vert_x, vert_y)
    vertices = [(vert_x, vert_y)] if not outline else []

    for side in range(sides+1):
        if (8<=side<=16):
            cosine = tl * cos(side * 2 * pi / sides) + vert_x
            sine = tl * sin(side * 2 * pi / sides) + vert_y
            vertices.append((cosine,sine))

    if not outline:
        batch = batch_for_shader(shader, draw_type, {"pos": vertices})
        shader.bind()
        shader.uniform_float("color", color)
        batch.draw(shader)
    else:
        batch = batch_for_shader(shader, draw_type, {"pos": [(v[0], v[1], 0) for v in vertices], "color": [color for v in vertices]})
        shader.bind()
        batch.draw(shader)

    # top right corner
    vert_x = area["vert"][0] + area["width"] - tr
    vert_y = area["vert"][1] - tr
    tr_vert = (vert_x, vert_y)
    vertices = [(vert_x, vert_y)] if not outline else []

    for side in range(sides+1):
        if (0<=side<=8):
            cosine = tr * cos(side * 2 * pi / sides) + vert_x
            sine = tr * sin(side * 2 * pi / sides) + vert_y
            vertices.append((cosine,sine))

    if not outline:
        batch = batch_for_shader(shader, draw_type, {"pos": vertices})
        shader.bind()
        shader.uniform_float("color", color)
        batch.draw(shader)
    else:
        batch = batch_for_shader(shader, draw_type, {"pos": [(v[0], v[1], 0) for v in vertices], "color": [color for v in vertices]})
        shader.bind()
        batch.draw(shader)

    # bottom left corner
    vert_x = area["vert"][0] + bl
    vert_y = area["vert"][1] - area["height"] + bl
    bl_vert = (vert_x, vert_y)
    vertices = [(vert_x, vert_y)] if not outline else []

    for side in range(sides+1):
        if (16<=side<=24):
            cosine = bl * cos(side * 2 * pi / sides) + vert_x
            sine = bl * sin(side * 2 * pi / sides) + vert_y
            vertices.append((cosine,sine))

    if not outline:
        batch = batch_for_shader(shader, draw_type, {"pos": vertices})
        shader.bind()
        shader.uniform_float("color", color)
        batch.draw(shader)
    else:
        batch = batch_for_shader(shader, draw_type, {"pos": [(v[0], v[1], 0) for v in vertices], "color": [color for v in vertices]})
        shader.bind()
        batch.draw(shader)

    # bottom right corner
    vert_x = area["vert"][0] + area["width"] - br
    vert_y = area["vert"][1] - area["height"] + br
    br_vert = (vert_x, vert_y)
    vertices = [(vert_x, vert_y)] if not outline else []

    for side in range(sides+1):
        if (24<=side<=32):
            cosine = br * cos(side * 2 * pi / sides) + vert_x
            sine = br * sin(side * 2 * pi / sides) + vert_y
            vertices.append((cosine,sine))

    if not outline:
        batch = batch_for_shader(shader, draw_type, {"pos": vertices})
        shader.bind()
        shader.uniform_float("color", color)
        batch.draw(shader)
    else:
        batch = batch_for_shader(shader, draw_type, {"pos": [(v[0], v[1], 0) for v in vertices], "color": [color for v in vertices]})
        shader.bind()
        batch.draw(shader)

    if not outline:
        vertices = []
        indices = []
        base_ind = 0

        # left edge
        width = max(tl, bl)
        le_x = tl_vert[0]-tl
        vertices.extend([
            (le_x, tl_vert[1]),
            (le_x+width, tl_vert[1]),
            (le_x, bl_vert[1]),
            (le_x+width, bl_vert[1])
            ])
        indices.extend([
            (base_ind,base_ind+1,base_ind+2),
            (base_ind+2,base_ind+3,base_ind+1)
            ])
        base_ind += 4

        # right edge
        width = max(tr, br)
        re_x = tr_vert[0]+tr
        vertices.extend([
            (re_x, tr_vert[1]),
            (re_x-width, tr_vert[1]),
            (re_x, br_vert[1]),
            (re_x-width, br_vert[1])
            ])
        indices.extend([
            (base_ind,base_ind+1,base_ind+2),
            (base_ind+2,base_ind+3,base_ind+1)
            ])
        base_ind += 4

        # top edge
        width = max(tl, tr)
        te_y = tl_vert[1]+tl
        vertices.extend([
            (tl_vert[0], te_y),
            (tl_vert[0], te_y-width),
            (tr_vert[0], te_y),
            (tr_vert[0], te_y-width)
            ])
        indices.extend([
            (base_ind,base_ind+1,base_ind+2),
            (base_ind+2,base_ind+3,base_ind+1)
            ])
        base_ind += 4

        # bottom edge
        width = max(bl, br)
        be_y = bl_vert[1]-bl
        vertices.extend([
            (bl_vert[0], be_y),
            (bl_vert[0], be_y+width),
            (br_vert[0], be_y),
            (br_vert[0], be_y+width)
            ])
        indices.extend([
            (base_ind,base_ind+1,base_ind+2),
            (base_ind+2,base_ind+3,base_ind+1)
            ])
        base_ind += 4

        # middle
        vertices.extend([
            tl_vert,
            tr_vert,
            bl_vert,
            br_vert
            ])
        indices.extend([
            (base_ind,base_ind+1,base_ind+2),
            (base_ind+2,base_ind+3,base_ind+1)
            ])

        batch = batch_for_shader(shader, 'TRIS', {"pos": vertices}, indices=indices)
        shader.bind()

        shader.uniform_float("color", color)
        batch.draw(shader)

    else:
        overlap = round(thickness / 2 - scale_factor() / 2)

        # left edge
        le_x = tl_vert[0]-tl
        vertices = [
            (le_x, tl_vert[1] + (overlap if tl == 0 else 0)),
            (le_x, bl_vert[1] - (overlap if bl == 0 else 0))
            ]

        batch = batch_for_shader(shader, 'LINE_STRIP', {"pos": [(v[0], v[1], 0) for v in vertices], "color": [color for v in vertices]})
        shader.bind()
        batch.draw(shader)

        # right edge
        re_x = tr_vert[0]+tr
        vertices = [
            (re_x, tr_vert[1] + (overlap if tr == 0 else 0)),
            (re_x, br_vert[1] - (overlap if br == 0 else 0))
            ]

        batch = batch_for_shader(shader, 'LINE_STRIP', {"pos": [(v[0], v[1], 0) for v in vertices], "color": [color for v in vertices]})
        shader.bind()
        batch.draw(shader)

        # top edge
        te_y = tl_vert[1]+tl
        vertices = [
            (tl_vert[0] - (overlap if tl == 0 else 0), te_y),
            (tr_vert[0] + (overlap if tr == 0 else 0), te_y)
            ]

        batch = batch_for_shader(shader, 'LINE_STRIP', {"pos": [(v[0], v[1], 0) for v in vertices], "color": [color for v in vertices]})
        shader.bind()
        batch.draw(shader)

        # bottom edge
        be_y = bl_vert[1]-bl
        vertices = [
            (bl_vert[0] - (overlap if bl == 0 else 0), be_y),
            (br_vert[0] + (overlap if br == 0 else 0), be_y)
            ]

        batch = batch_for_shader(shader, 'LINE_STRIP', {"pos": [(v[0], v[1], 0) for v in vertices], "color": [color for v in vertices]})
        shader.bind()
        batch.draw(shader)


    gpu.state.blend_set('NONE')

def mouse_in_area(mouse_pos, area, buf = 0):
    x = mouse_pos[0]
    y = mouse_pos[1]

    # check left
    if x+buf < area["vert"][0]:
        return False

    # check right
    if x-buf > area["vert"][0] + area["width"]:
        return False

    # check top
    if y-buf > area["vert"][1]:
        return False

    # check bottom
    if y+buf < area["vert"][1] - area["height"]:
        return False

    # if we reach here we're in the area
    return True

def account_for_view_bounds(area):
    # make sure it renders in the 3d view - prioritize top left

    # right
    if area["vert"][0] + area["width"] > bpy.context.region.width:
        x = bpy.context.region.width - area["width"]
        y = area["vert"][1]

        area["vert"] = (x, y)

    # left
    if area["vert"][0] < 0:
        x = 0
        y = area["vert"][1]

        area["vert"] = (x, y)

    # bottom
    if area["vert"][1] - area["height"] < 0:
        x = area["vert"][0]
        y = area["height"]

        area["vert"] = (x, y)

    # top
    if area["vert"][1] > bpy.context.region.height:
        x = area["vert"][0]
        y = bpy.context.region.height

        area["vert"] = (x, y)

def update_area_dimensions(area, w=0, h=0):
    area["width"] += w
    area["height"] += h

class QCDMoveWidget(Operator):
    """Move objects to QCD Slots"""
    bl_idname = "view3d.qcd_move_widget"
    bl_label = "QCD Move Widget"

    slots = {
        "ONE":1,
        "TWO":2,
        "THREE":3,
        "FOUR":4,
        "FIVE":5,
        "SIX":6,
        "SEVEN":7,
        "EIGHT":8,
        "NINE":9,
        "ZERO":10,
        }

    last_type = ''
    last_type_value = ''
    initialized = False
    moved = False

    def modal(self, context, event):
        if event.type == 'TIMER':
            if self.hover_time and self.hover_time + 0.5 < time.time():
                self.draw_tooltip = True

                context.area.tag_redraw()
            return {'RUNNING_MODAL'}


        context.area.tag_redraw()

        if len(self.areas) == 1:
            return {'RUNNING_MODAL'}

        if self.last_type == 'LEFTMOUSE' and self.last_type_value == 'PRESS' and event.type == 'MOUSEMOVE':
            if mouse_in_area(self.mouse_pos, self.areas["Grab Bar"]):
                x_offset = self.areas["Main Window"]["vert"][0] - self.mouse_pos[0]
                x = event.mouse_region_x + x_offset

                y_offset = self.areas["Main Window"]["vert"][1] - self.mouse_pos[1]
                y = event.mouse_region_y + y_offset

                self.areas["Main Window"]["vert"] = (x, y)

                self.mouse_pos = (event.mouse_region_x, event.mouse_region_y)

        elif event.type == 'MOUSEMOVE':
            self.draw_tooltip = False
            self.hover_time = None
            self.mouse_pos = (event.mouse_region_x, event.mouse_region_y)

            if not mouse_in_area(self.mouse_pos, self.areas["Main Window"], 50 * scale_factor()):
                if self.initialized:
                    bpy.types.SpaceView3D.draw_handler_remove(self._handle, 'WINDOW')

                    if self.moved:
                        bpy.ops.ed.undo_push()

                    return {'FINISHED'}

            else:
                self.initialized = True

        elif event.value == 'PRESS' and event.type == 'LEFTMOUSE':
            if not mouse_in_area(self.mouse_pos, self.areas["Main Window"], 10 * scale_factor()):
                bpy.types.SpaceView3D.draw_handler_remove(self._handle, 'WINDOW')

                if self.moved:
                    bpy.ops.ed.undo_push()

                return {'FINISHED'}

            for num in range(20):
                if not self.areas.get(f"Button {num + 1}", None):
                    continue

                if mouse_in_area(self.mouse_pos, self.areas[f"Button {num + 1}"]):
                    bpy.ops.view3d.move_to_qcd_slot(slot=str(num + 1), toggle=event.shift)
                    self.moved = True

        elif event.type in {'RIGHTMOUSE', 'ESC'}:
            bpy.types.SpaceView3D.draw_handler_remove(self._handle, 'WINDOW')

            return {'CANCELLED'}

        if event.value == 'PRESS' and event.type in self.slots:
            move_to = self.slots[event.type]

            if event.alt:
                move_to += 10

            if event.shift:
                bpy.ops.view3d.move_to_qcd_slot(slot=str(move_to), toggle=True)
            else:
                bpy.ops.view3d.move_to_qcd_slot(slot=str(move_to), toggle=False)

            self.moved = True

        if event.type != 'MOUSEMOVE' and event.type != 'INBETWEEN_MOUSEMOVE':
            self.last_type = event.type
            self.last_type_value = event.value

        return {'RUNNING_MODAL'}

    def invoke(self, context, event):
        if context.area.type == 'VIEW_3D':
            # the arguments we pass the the callback
            args = (self, context)
            # Add the region OpenGL drawing callback
            # draw in view space with 'POST_VIEW' and 'PRE_VIEW'
            self._handle = bpy.types.SpaceView3D.draw_handler_add(draw_callback_px, args, 'WINDOW', 'POST_PIXEL')
            self._timer = context.window_manager.event_timer_add(0.1, window=context.window)

            self.mouse_pos = (event.mouse_region_x, event.mouse_region_y)

            self.draw_tooltip = False

            self.hover_time = None

            self.areas = {}

            # MAIN WINDOW BACKGROUND
            x = self.mouse_pos[0] - spacer()*2
            y = self.mouse_pos[1] + spacer()*2
            main_window = {
                # Top Left Vertex
                "vert": (x,y),
                "width": 0,
                "height": 0,
                "value": None
                }

            self.areas["Main Window"] = main_window
            allocate_main_ui(self, context)
            account_for_view_bounds(main_window)

            context.window_manager.modal_handler_add(self)
            return {'RUNNING_MODAL'}

        else:
            self.report({'WARNING'}, "View3D not found, cannot run operator")
            return {'CANCELLED'}


def allocate_main_ui(self, context):
    main_window = self.areas["Main Window"]
    self.areas.clear()
    main_window["width"] = 0
    main_window["height"] = 0
    self.areas["Main Window"] = main_window

    cur_width_pos = main_window["vert"][0]
    cur_height_pos = main_window["vert"][1]

    # GRAB BAR
    grab_bar = {
        "vert": main_window["vert"],
        "width": 0,
        "height": round(23 * scale_factor()),
        "value": None
        }

    # add grab bar to areas
    self.areas["Grab Bar"] = grab_bar


    # WINDOW TITLE
    wt_indent_x = spacer()*2
    wt_y_offset = round(spacer()/2)
    window_title = {
        "vert": main_window["vert"],
        "width": 0,
        "height": round(13 * scale_factor()),
        "value": "Move Objects to QCD Slots"
        }

    x = main_window["vert"][0] + wt_indent_x
    y = main_window["vert"][1] - window_title["height"] - wt_y_offset
    window_title["vert"] = (x, y)

    # add window title to areas
    self.areas["Window Title"] = window_title

    cur_height_pos = window_title["vert"][1]


    # MAIN BUTTON AREA
    button_size = round(20 * scale_factor())
    button_gap = round(1 * scale_factor())
    button_group = 5
    button_group_gap = round(20 * scale_factor())
    button_group_width = button_size * button_group + button_gap * (button_group - 1)

    mba_indent_x = spacer()*2
    mba_outdent_x = spacer()*2
    mba_indent_y = spacer()
    x = cur_width_pos + mba_indent_x
    y = cur_height_pos - mba_indent_y
    main_button_area = {
        "vert": (x, y),
        "width": 0,
        "height": 0,
        "value": None
        }

    # add main button area to areas
    self.areas["Main Button Area"] = main_button_area

    # update current position
    cur_width_pos = main_button_area["vert"][0]
    cur_height_pos = main_button_area["vert"][1]


    # BUTTON ROW 1 A
    button_row_1_a = {
        "vert": main_button_area["vert"],
        "width": button_group_width,
        "height": button_size,
        "value": None
        }

    # add button row 1 A to areas
    self.areas["Button Row 1 A"] = button_row_1_a

    # advance width pos to start of next row
    cur_width_pos += button_row_1_a["width"]
    cur_width_pos += button_group_gap

    # BUTTON ROW 1 B
    x = cur_width_pos
    y = cur_height_pos
    button_row_1_b = {
        "vert": (x, y),
        "width": button_group_width,
        "height": button_size,
        "value": None
        }

    # add button row 1 B to areas
    self.areas["Button Row 1 B"] = button_row_1_b

    # reset width pos to start of main button area
    cur_width_pos = main_button_area["vert"][0]
    # update height pos
    cur_height_pos -= button_row_1_a["height"]
    # add gap between button rows
    cur_height_pos -= button_gap


    # BUTTON ROW 2 A
    x = cur_width_pos
    y = cur_height_pos
    button_row_2_a = {
        "vert": (x, y),
        "width": button_group_width,
        "height": button_size,
        "value": None
        }

    # add button row 2 A to areas
    self.areas["Button Row 2 A"] = button_row_2_a

    # advance width pos to start of next row
    cur_width_pos += button_row_2_a["width"]
    cur_width_pos += button_group_gap

    # BUTTON ROW 2 B
    x = cur_width_pos
    y = cur_height_pos
    button_row_2_b = {
        "vert": (x, y),
        "width": button_group_width,
        "height": button_size,
        "value": None
        }

    # add button row 2 B to areas
    self.areas["Button Row 2 B"] = button_row_2_b


    selected_objects = get_move_selection()
    active_object = get_move_active()


    # BUTTONS
    def get_buttons(button_row, row_num):
        cur_width_pos = button_row["vert"][0]
        cur_height_pos = button_row["vert"][1]
        for num in range(button_group):
            slot_num = row_num + num

            qcd_slot_name = internals.qcd_slots.get_name(f"{slot_num}")

            if qcd_slot_name:
                qcd_laycol = internals.layer_collections[qcd_slot_name]["ptr"]
                collection_objects = qcd_laycol.collection.objects

                # BUTTON
                x = cur_width_pos
                y = cur_height_pos
                button = {
                    "vert": (x, y),
                    "width": button_size,
                    "height": button_size,
                    "value": slot_num
                    }

                self.areas[f"Button {slot_num}"] = button

                # ACTIVE OBJECT ICON
                if active_object and active_object in selected_objects and active_object.name in collection_objects:
                    x = cur_width_pos + round(button_size / 4)
                    y = cur_height_pos - round(button_size / 4)
                    active_object_indicator = {
                    "vert": (x, y),
                    "width": floor(button_size / 2),
                    "height": floor(button_size / 2),
                    "value": None
                    }

                    self.areas[f"Button {slot_num} Active Object Indicator"] = active_object_indicator

                elif not set(selected_objects).isdisjoint(collection_objects):
                    x = cur_width_pos + round(button_size / 4) + floor(1 * scale_factor())
                    y = cur_height_pos - round(button_size / 4) - floor(1 * scale_factor())
                    selected_object_indicator = {
                        "vert": (x, y),
                        "width": floor(button_size / 2) - floor(1 * scale_factor()),
                        "height": floor(button_size / 2) - floor(1 * scale_factor()),
                        "value": None
                        }

                    self.areas[f"Button {slot_num} Selected Object Indicator"] = selected_object_indicator

                elif collection_objects:
                    x = cur_width_pos + floor(button_size / 4)
                    y = cur_height_pos - button_size / 2 + 1 * scale_factor()
                    object_indicator = {
                    "vert": (x, y),
                    "width": round(button_size / 2),
                    "height": round(2 * scale_factor()),
                    "value": None
                    }
                    self.areas[f"Button {slot_num} Object Indicator"] = object_indicator

            else:
                x = cur_width_pos + 2 * scale_factor()
                y = cur_height_pos - 2 * scale_factor()
                X_icon = {
                    "vert": (x, y),
                    "width": button_size - 4 * scale_factor(),
                    "height": button_size - 4 * scale_factor(),
                    "value": None
                    }

                self.areas[f"X_icon {slot_num}"] = X_icon

            cur_width_pos += button_size
            cur_width_pos += button_gap

    get_buttons(button_row_1_a, 1)
    get_buttons(button_row_1_b, 6)
    get_buttons(button_row_2_a, 11)
    get_buttons(button_row_2_b, 16)


    # UPDATE DYNAMIC DIMENSIONS
    width = button_row_1_a["width"] + button_group_gap + button_row_1_b["width"]
    height = button_row_1_a["height"] + button_gap + button_row_2_a["height"]
    update_area_dimensions(main_button_area, width, height)

    width = main_button_area["width"] + mba_indent_x + mba_outdent_x
    height = main_button_area["height"] + mba_indent_y * 2 + window_title["height"] + wt_y_offset
    update_area_dimensions(main_window, width, height)

    update_area_dimensions(grab_bar, main_window["width"])


def draw_callback_px(self, context):
    allocate_main_ui(self, context)

    shader = gpu.shader.from_builtin('UNIFORM_COLOR')
    line_shader = gpu.shader.from_builtin('POLYLINE_SMOOTH_COLOR')

    addon_prefs = context.preferences.addons[__package__].preferences

    # main window background
    main_window = self.areas["Main Window"]
    outline_color = addon_prefs.qcd_ogl_widget_menu_back_outline
    background_color = addon_prefs.qcd_ogl_widget_menu_back_inner
    draw_rounded_rect(main_window, line_shader, outline_color[:], outline=True)
    draw_rounded_rect(main_window, shader, background_color)

    # draw window title
    window_title = self.areas["Window Title"]
    x = window_title["vert"][0]
    y = window_title["vert"][1]
    h = window_title["height"]
    text = window_title["value"]
    text_color = addon_prefs.qcd_ogl_widget_menu_back_text
    font_id = 0
    blf.position(font_id, x, y, 0)
    blf.size(font_id, int(h))
    blf.color(font_id, text_color[0], text_color[1], text_color[2], 1)
    blf.draw(font_id, text)

    in_tooltip_area = False
    tooltip_slot_idx = None

    for num in range(20):
        slot_num = num + 1
        qcd_slot_name = internals.qcd_slots.get_name(f"{slot_num}")
        if qcd_slot_name:
            qcd_laycol = internals.layer_collections[qcd_slot_name]["ptr"]
            collection_objects = qcd_laycol.collection.objects
            selected_objects = get_move_selection()
            active_object = get_move_active()
            button_area = self.areas[f"Button {slot_num}"]

            # colors
            button_color = addon_prefs.qcd_ogl_widget_tool_inner
            icon_color = addon_prefs.qcd_ogl_widget_tool_text
            if not qcd_laycol.exclude:
                button_color = addon_prefs.qcd_ogl_widget_tool_inner_sel
                icon_color = addon_prefs.qcd_ogl_widget_tool_text_sel

            if mouse_in_area(self.mouse_pos, button_area):
                in_tooltip_area = True
                tooltip_slot_idx = slot_num

                mod = 0.1

                if button_color[0] + mod > 1 or button_color[1] + mod > 1 or button_color[2] + mod > 1:
                    mod = -mod

                button_color = (
                    button_color[0] + mod,
                    button_color[1] + mod,
                    button_color[2] + mod,
                    button_color[3]
                    )


            # button roundness
            tl = tr = bl = br = 0
            rounding = 5

            if num < 10:
                if not internals.qcd_slots.contains(idx=f"{num+2}"):
                    tr = rounding

                if not internals.qcd_slots.contains(idx=f"{num}"):
                    tl = rounding
            else:
                if not internals.qcd_slots.contains(idx=f"{num+2}"):
                    br = rounding

                if not internals.qcd_slots.contains(idx=f"{num}"):
                    bl = rounding

            if num in [0,5]:
                tl = rounding
            elif num in [4,9]:
                tr = rounding
            elif num in [10,15]:
                bl = rounding
            elif num in [14,19]:
                br = rounding

            # draw button
            outline_color = addon_prefs.qcd_ogl_widget_tool_outline
            draw_rounded_rect(button_area, line_shader, outline_color[:], tl, tr, bl, br, outline=True)
            draw_rounded_rect(button_area, shader, button_color, tl, tr, bl, br)

            # ACTIVE OBJECT
            if active_object and active_object in selected_objects and active_object.name in collection_objects:
                active_object_indicator = self.areas[f"Button {slot_num} Active Object Indicator"]

                vertices = get_circle_coords(active_object_indicator)
                batch = batch_for_shader(shader, 'TRI_FAN', {"pos": vertices})
                shader.bind()
                shader.uniform_float("color", icon_color[:] + (1,))

                gpu.state.blend_set('ALPHA')

                batch.draw(shader)

                gpu.state.blend_set('NONE')

            # SELECTED OBJECTS
            elif not set(selected_objects).isdisjoint(collection_objects):
                selected_object_indicator = self.areas[f"Button {slot_num} Selected Object Indicator"]

                alpha = addon_prefs.qcd_ogl_selected_icon_alpha
                vertices = get_circle_coords(selected_object_indicator)
                line_shader.uniform_float("lineWidth", 2 * scale_factor())
                color = icon_color[:] + (alpha,)
                batch = batch_for_shader(line_shader, 'LINE_STRIP', {"pos": [(v[0], v[1], 0) for v in vertices], "color": [color for v in vertices]})
                shader.bind()

                gpu.state.blend_set('ALPHA')

                batch.draw(line_shader)

                gpu.state.blend_set('NONE')

            # OBJECTS
            elif collection_objects:
                object_indicator = self.areas[f"Button {slot_num} Object Indicator"]

                alpha = addon_prefs.qcd_ogl_objects_icon_alpha
                vertices, indices = get_coords(object_indicator)
                batch = batch_for_shader(shader, 'TRIS', {"pos": vertices}, indices=indices)
                shader.bind()
                shader.uniform_float("color", icon_color[:] + (alpha,))

                gpu.state.blend_set('ALPHA')

                batch.draw(shader)

                gpu.state.blend_set('NONE')


        # X ICON
        else:
            X_icon = self.areas[f"X_icon {slot_num}"]
            X_icon_color = addon_prefs.qcd_ogl_widget_menu_back_text

            vertices, indices = get_x_coords(X_icon)
            batch = batch_for_shader(shader, 'TRIS', {"pos": vertices}, indices=indices)
            shader.bind()
            shader.uniform_float("color", X_icon_color[:] + (1,))

            gpu.state.blend_set('ALPHA')

            batch.draw(shader)

            gpu.state.blend_set('NONE')

    if in_tooltip_area:
        if self.draw_tooltip:
            slot_name = internals.qcd_slots.get_name(f"{tooltip_slot_idx}")
            slot_string = f"QCD Slot {tooltip_slot_idx}: \"{slot_name}\"\n"
            hotkey_string = (
                "  * LMB - Move objects to slot.\n"
                "  * Shift+LMB - Toggle objects\' slot."
                )

            draw_tooltip(self, context, shader, line_shader, f"{slot_string}{hotkey_string}")

            self.hover_time = None

        else:
            if not self.hover_time:
                self.hover_time = time.time()


def draw_tooltip(self, context, shader, line_shader, message):
    addon_prefs = context.preferences.addons[__package__].preferences

    font_id = 0
    line_height = 11 * scale_factor()
    text_color = addon_prefs.qcd_ogl_widget_tooltip_text
    blf.size(font_id, int(line_height))
    blf.color(font_id, text_color[0], text_color[1], text_color[2], 1)

    lines = message.split("\n")
    longest = [0,""]
    num_lines = len(lines)

    for line in lines:
        w, _ = blf.dimensions(font_id, line)

        if w > longest[0]:
            longest[0] = w
            longest[1] = line

    w, h = blf.dimensions(font_id, longest[1])

    line_spacer = 1 * scale_factor()
    padding = 4 * scale_factor()

    # draw background
    tooltip = {
        "vert": self.mouse_pos,
        "width": w + spacer()*2,
        "height": (line_height * num_lines + line_spacer * num_lines) + padding*3,
        "value": None
        }

    x = tooltip["vert"][0] - spacer()*2
    y = tooltip["vert"][1] + tooltip["height"] + round(5 * scale_factor())
    tooltip["vert"] = (x, y)

    account_for_view_bounds(tooltip)

    outline_color = addon_prefs.qcd_ogl_widget_tooltip_outline
    background_color = addon_prefs.qcd_ogl_widget_tooltip_inner
    draw_rounded_rect(tooltip, line_shader, outline_color[:], outline=True)
    draw_rounded_rect(tooltip, shader, background_color)

    line_pos = padding + line_height
    # draw text
    for num, line in enumerate(lines):
        x = tooltip["vert"][0] + spacer()
        y = tooltip["vert"][1] - line_pos
        blf.position(font_id, x, y, 0)
        blf.draw(font_id, line)

        line_pos += line_height + line_spacer



# SPDX-FileCopyrightText: 2020-2023 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later

import bpy

class GP_PT_sidebarPanel(bpy.types.Panel):
    bl_label = "Grease Pencil Tools"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_category = "Grease Pencil"

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        # Box deform ops
        self.layout.operator_context = 'INVOKE_DEFAULT'
        layout.operator('view3d.gp_box_deform', icon ="MOD_MESHDEFORM")

        # Straight line ops
        layout.operator('gpencil.straight_stroke', icon ="CURVE_PATH")


        # Expose native view operators
        row = layout.row(align=True)
        row.operator('view3d.zoom_camera_1_to_1', text = 'Zoom 1:1', icon = 'ZOOM_PREVIOUS')
        row.operator('view3d.view_center_camera', text = 'Zoom Fit', icon = 'FULLSCREEN_ENTER')

        # Rotation save/load
        row = layout.row(align=True)
        row.operator('view3d.rotate_canvas_reset', text = 'Reset Rotation', icon = 'FILE_REFRESH')
        row.operator('view3d.rotate_canvas_set', text = 'Save Rotation', icon = 'DRIVER_ROTATIONAL_DIFFERENCE')

        # View flip
        if context.scene.camera and context.scene.camera.scale.x < 0:
            row = layout.row(align=True)
            row.operator('view3d.camera_flip_x', text = 'Camera Mirror Flip', icon = 'MOD_MIRROR')
            row.label(text='', icon='LOOP_BACK')
        else:
            layout.operator('view3d.camera_flip_x', text = 'Camera Mirror Flip', icon = 'MOD_MIRROR')


def menu_boxdeform_entry(self, context):
    """Transform shortcut to append in existing menu"""
    layout = self.layout
    obj = bpy.context.object
    # {'EDIT_GPENCIL', 'PAINT_GPENCIL','SCULPT_GPENCIL','WEIGHT_GPENCIL', 'VERTEX_GPENCIL'}
    if obj and obj.type == 'GPENCIL' and context.mode in {'OBJECT', 'EDIT_GPENCIL', 'PAINT_GPENCIL'}:
        self.layout.operator_context = 'INVOKE_DEFAULT'
        layout.operator('view3d.gp_box_deform', text='Box Deform')

def menu_stroke_entry(self, context):
    layout = self.layout
    # Gpencil modes : {'EDIT_GPENCIL', 'PAINT_GPENCIL','SCULPT_GPENCIL','WEIGHT_GPENCIL', 'VERTEX_GPENCIL'}
    if context.mode in {'EDIT_GPENCIL', 'PAINT_GPENCIL'}:
        self.layout.operator_context = 'INVOKE_DEFAULT'
        layout.operator('gpencil.straight_stroke', text='Straight Stroke')

def menu_brush_pack(self, context):
    layout = self.layout
    # if context.mode in {'EDIT_GPENCIL', 'PAINT_GPENCIL'}:
    self.layout.operator_context = 'INVOKE_DEFAULT'
    layout.operator('gp.import_brush_pack')#, text='Import brush pack'


def register():
    bpy.utils.register_class(GP_PT_sidebarPanel)
    ## VIEW3D_MT_edit_gpencil.append# Grease pencil menu
    bpy.types.VIEW3D_MT_transform_object.append(menu_boxdeform_entry)
    bpy.types.VIEW3D_MT_edit_gpencil_transform.append(menu_boxdeform_entry)
    bpy.types.VIEW3D_MT_edit_gpencil_stroke.append(menu_stroke_entry)
    bpy.types.VIEW3D_MT_brush_gpencil_context_menu.append(menu_brush_pack)


def unregister():
    bpy.types.VIEW3D_MT_brush_gpencil_context_menu.remove(menu_brush_pack)
    bpy.types.VIEW3D_MT_transform_object.remove(menu_boxdeform_entry)
    bpy.types.VIEW3D_MT_edit_gpencil_transform.remove(menu_boxdeform_entry)
    bpy.types.VIEW3D_MT_edit_gpencil_stroke.remove(menu_stroke_entry)
    bpy.utils.unregister_class(GP_PT_sidebarPanel)



# SPDX-FileCopyrightText: 2016-2020 by Nathan Lovato, Daniel Oakey, Razvan Radulescu, and contributors
#
# SPDX-License-Identifier: GPL-3.0-or-later

import bpy

from .utils.doc import doc_name, doc_idname, doc_brief, doc_description


class POWER_SEQUENCER_OT_markers_create_from_selected_strips(bpy.types.Operator):
    """
    *brief* Create one marker at the start on each selected strip, based on its name

    Use it to copy markers as timecodes.
    """

    doc = {
        "name": doc_name(__qualname__),
        "demo": "",
        "description": doc_description(__doc__),
        "shortcuts": [],
        "keymap": "Sequencer",
    }
    bl_idname = doc_idname(__qualname__)
    bl_label = doc["name"]
    bl_description = doc_brief(doc["description"])

    @classmethod
    def poll(cls, context):
        return context.selected_sequences

    def execute(self, context):
        cursor_frame_start = context.scene.frame_current

        for m in context.scene.timeline_markers:
            m.select = False

        for s in context.selected_sequences:
            bpy.ops.marker.add()
            new_marker = context.scene.timeline_markers[-1]

            new_marker.select = True
            bpy.ops.marker.rename(name=s.name)
            gap = s.frame_final_start - cursor_frame_start
            bpy.ops.marker.move(frames=gap)
            new_marker.select = False
        return {"FINISHED"}



# SPDX-FileCopyrightText: 2019-2022 Alan Odom (Clockmender)
# SPDX-FileCopyrightText: 2019-2022 Rune Morling (ermo)
#
# SPDX-License-Identifier: GPL-2.0-or-later

# Common Functions used in more than one place in PDT Operations

import bpy
import bmesh
import gpu
import numpy as np
from mathutils import Vector, Quaternion
from gpu_extras.batch import batch_for_shader
from math import cos, sin, pi
from .pdt_msg_strings import (
    PDT_ERR_VERT_MODE,
    PDT_ERR_SEL_2_V_1_E,
    PDT_ERR_SEL_2_OBJS,
    PDT_ERR_NO_ACT_OBJ,
    PDT_ERR_SEL_1_EDGEM,
)
from . import pdt_exception
PDT_ShaderError = pdt_exception.ShaderError


def debug(msg, prefix=""):
    """Print a debug message to the console if PDT's or Blender's debug flags are set.

    Note:
        The printed message will be of the form:

        {prefix}{caller file name:line number}| {msg}

    Args:
        msg: Incoming message to display
        prefix: Always Blank

    Returns:
        Nothing.
    """

    pdt_debug = bpy.context.preferences.addons[__package__].preferences.debug
    if  bpy.app.debug or bpy.app.debug_python or pdt_debug:
        import traceback

        def extract_filename(fullpath):
            """Return only the filename part of fullpath (excluding its path).

            Args:
                fullpath: Filename's full path

            Returns:
                filename.
            """
            # Expected to end up being a string containing only the filename
            # (i.e. excluding its preceding '/' separated path)
            filename = fullpath.split('/')[-1]
            #print(filename)
            # something went wrong
            if len(filename) < 1:
                return fullpath
            # since this is a string, just return it
            return filename

        # stack frame corresponding to the line where debug(msg) was called
        #print(traceback.extract_stack()[-2])
        laststack = traceback.extract_stack()[-2]
        #print(laststack[0])
        # laststack[0] is the caller's full file name, laststack[1] is the line number
        print(f"{prefix}{extract_filename(laststack[0])}:{laststack[1]}| {msg}")

def oops(self, context):
    """Error Routine.

    Note:
        Displays error message in a popup.

    Args:
        context: Blender bpy.context instance.

    Returns:
        Nothing.
    """

    scene = context.scene
    pg = scene.pdt_pg
    self.layout.label(text=pg.error)


def set_mode(mode_pl):
    """Sets Active Axes for View Orientation.

    Note:
        Sets indices of axes for locational vectors:
        a3 is normal to screen, or depth
        "XY": a1 = x, a2 = y, a3 = z
        "XZ": a1 = x, a2 = z, a3 = y
        "YZ": a1 = y, a2 = z, a3 = x

    Args:
        mode_pl: Plane Selector variable as input

    Returns:
        3 Integer indices.
    """

    order = {
        "XY": (0, 1, 2),
        "XZ": (0, 2, 1),
        "YZ": (1, 2, 0),
        "LO": (0, 1, 2),
    }
    return order[mode_pl]


def set_axis(mode_pl):
    """Sets Active Axes for View Orientation.

    Note:
        Sets indices for axes from taper vectors
        Axis order: Rotate Axis, Move Axis, Height Axis

    Args:
        mode_pl: Taper Axis Selector variable as input

    Returns:
        3 Integer Indices.
    """

    order = {
        "RX-MY": (0, 1, 2),
        "RX-MZ": (0, 2, 1),
        "RY-MX": (1, 0, 2),
        "RY-MZ": (1, 2, 0),
        "RZ-MX": (2, 0, 1),
        "RZ-MY": (2, 1, 0),
    }
    return order[mode_pl]


def check_selection(num, bm, obj):
    """Check that the Object's select_history has sufficient entries.

    Note:
        If selection history is not Verts, clears selection and history.

    Args:
        num: The number of entries required for each operation
        bm: The Bmesh from the Object
        obj: The Object

    Returns:
        list of 3D points as Vectors.
    """

    if len(bm.select_history) < num:
        return None
    active_vertex = bm.select_history[-1]
    if isinstance(active_vertex, bmesh.types.BMVert):
        vector_a = active_vertex.co
        if num == 1:
            return vector_a
        if num == 2:
            vector_b = bm.select_history[-2].co
            return vector_a, vector_b
        if num == 3:
            vector_b = bm.select_history[-2].co
            vector_c = bm.select_history[-3].co
            return vector_a, vector_b, vector_c
        if num == 4:
            vector_b = bm.select_history[-2].co
            vector_c = bm.select_history[-3].co
            vector_d = bm.select_history[-4].co
            return vector_a, vector_b, vector_c, vector_d
    else:
        for f in bm.faces:
            f.select_set(False)
        for e in bm.edges:
            e.select_set(False)
        for v in bm.verts:
            v.select_set(False)
        bmesh.update_edit_mesh(obj.data)
        bm.select_history.clear()
    return None


def update_sel(bm, verts, edges, faces):
    """Updates Vertex, Edge and Face Selections following a function.

    Args:
        bm: Object Bmesh
        verts: New Selection for Vertices
        edges: The Edges on which to operate
        faces: The Faces on which to operate

    Returns:
        Nothing.
    """
    for f in bm.faces:
        f.select_set(False)
    for e in bm.edges:
        e.select_set(False)
    for v in bm.verts:
        v.select_set(False)
    for v in verts:
        v.select_set(True)
    for e in edges:
        e.select_set(True)
    for f in faces:
        f.select_set(True)


def view_coords(x_loc, y_loc, z_loc):
    """Converts input Vector values to new Screen Oriented Vector.

    Args:
        x_loc: X coordinate from vector
        y_loc: Y coordinate from vector
        z_loc: Z coordinate from vector

    Returns:
        Vector adjusted to View's Inverted Transformation Matrix.
    """

    areas = [a for a in bpy.context.screen.areas if a.type == "VIEW_3D"]
    if len(areas) > 0:
        view_matrix = areas[0].spaces.active.region_3d.view_matrix
        view_matrix = view_matrix.to_3x3().normalized().inverted()
        view_location = Vector((x_loc, y_loc, z_loc))
        new_view_location = view_matrix @ view_location
        return new_view_location

    return Vector((0, 0, 0))


def view_coords_i(x_loc, y_loc, z_loc):
    """Converts Screen Oriented input Vector values to new World Vector.

    Note:
        Converts View transformation Matrix to Rotational Matrix

    Args:
        x_loc: X coordinate from vector
        y_loc: Y coordinate from vector
        z_loc: Z coordinate from vector

    Returns:
        Vector adjusted to View's Transformation Matrix.
    """

    areas = [a for a in bpy.context.screen.areas if a.type == "VIEW_3D"]
    if len(areas) > 0:
        view_matrix = areas[0].spaces.active.region_3d.view_matrix
        view_matrix = view_matrix.to_3x3().normalized()
        view_location = Vector((x_loc, y_loc, z_loc))
        new_view_location = view_matrix @ view_location
        return new_view_location

    return Vector((0, 0, 0))


def view_dir(dis_v, ang_v):
    """Converts Distance and Angle to View Oriented Vector.

    Note:
        Converts View Transformation Matrix to Rotational Matrix (3x3)
        Angles are Converts to Radians from degrees.

    Args:
        dis_v: Scene PDT distance
        ang_v: Scene PDT angle

    Returns:
        World Vector.
    """

    areas = [a for a in bpy.context.screen.areas if a.type == "VIEW_3D"]
    if len(areas) > 0:
        view_matrix = areas[0].spaces.active.region_3d.view_matrix
        view_matrix = view_matrix.to_3x3().normalized().inverted()
        view_location = Vector((0, 0, 0))
        view_location.x = dis_v * cos(ang_v * pi / 180)
        view_location.y = dis_v * sin(ang_v * pi / 180)
        new_view_location = view_matrix @ view_location
        return new_view_location

    return Vector((0, 0, 0))


def euler_to_quaternion(roll, pitch, yaw):
    """Converts Euler Rotation to Quaternion Rotation.

    Args:
        roll: Roll in Euler rotation
        pitch: Pitch in Euler rotation
        yaw: Yaw in Euler rotation

    Returns:
        Quaternion Rotation.
    """

    # fmt: off
    quat_x = (np.sin(roll/2) * np.cos(pitch/2) * np.cos(yaw/2)
              - np.cos(roll/2) * np.sin(pitch/2) * np.sin(yaw/2))
    quat_y = (np.cos(roll/2) * np.sin(pitch/2) * np.cos(yaw/2)
              + np.sin(roll/2) * np.cos(pitch/2) * np.sin(yaw/2))
    quat_z = (np.cos(roll/2) * np.cos(pitch/2) * np.sin(yaw/2)
              - np.sin(roll/2) * np.sin(pitch/2) * np.cos(yaw/2))
    quat_w = (np.cos(roll/2) * np.cos(pitch/2) * np.cos(yaw/2)
              + np.sin(roll/2) * np.sin(pitch/2) * np.sin(yaw/2))
    # fmt: on
    return Quaternion((quat_w, quat_x, quat_y, quat_z))


def arc_centre(vector_a, vector_b, vector_c):
    """Calculates Centre of Arc from 3 Vector Locations using standard Numpy routine

    Args:
        vector_a: Active vector location
        vector_b: Second vector location
        vector_c: Third vector location

    Returns:
        Vector representing Arc Centre and Float representing Arc Radius.
    """

    coord_a = np.array([vector_a.x, vector_a.y, vector_a.z])
    coord_b = np.array([vector_b.x, vector_b.y, vector_b.z])
    coord_c = np.array([vector_c.x, vector_c.y, vector_c.z])
    line_a = np.linalg.norm(coord_c - coord_b)
    line_b = np.linalg.norm(coord_c - coord_a)
    line_c = np.linalg.norm(coord_b - coord_a)
    # fmt: off
    line_s = (line_a+line_b+line_c) / 2
    radius = (
        line_a*line_b*line_c/4
        / np.sqrt(line_s
                  * (line_s-line_a)
                  * (line_s-line_b)
                  * (line_s-line_c))
        )
    base_1 = line_a*line_a * (line_b*line_b + line_c*line_c - line_a*line_a)
    base_2 = line_b*line_b * (line_a*line_a + line_c*line_c - line_b*line_b)
    base_3 = line_c*line_c * (line_a*line_a + line_b*line_b - line_c*line_c)
    # fmt: on
    intersect_coord = np.column_stack((coord_a, coord_b, coord_c))
    intersect_coord = intersect_coord.dot(np.hstack((base_1, base_2, base_3)))
    intersect_coord /= base_1 + base_2 + base_3
    return Vector((intersect_coord[0], intersect_coord[1], intersect_coord[2])), radius


def intersection(vertex_a, vertex_b, vertex_c, vertex_d, plane):
    """Calculates Intersection Point of 2 Imagined Lines from 4 Vectors.

    Note:
       Calculates Converging Intersect Location and indication of
       whether the lines are convergent using standard Numpy Routines

    Args:
        vertex_a: Active vector location of first line
        vertex_b: Second vector location of first line
        vertex_c: Third vector location of 2nd line
        vertex_d: Fourth vector location of 2nd line
        plane: Working Plane 4 Vector Locations representing 2 lines and Working Plane

    Returns:
        Intersection Vector and Boolean for convergent state.
    """

    if plane == "LO":
        vertex_offset = vertex_b - vertex_a
        vertex_b = view_coords_i(vertex_offset.x, vertex_offset.y, vertex_offset.z)
        vertex_offset = vertex_d - vertex_a
        vertex_d = view_coords_i(vertex_offset.x, vertex_offset.y, vertex_offset.z)
        vertex_offset = vertex_c - vertex_a
        vertex_c = view_coords_i(vertex_offset.x, vertex_offset.y, vertex_offset.z)
        vector_ref = Vector((0, 0, 0))
        coord_a = (vertex_c.x, vertex_c.y)
        coord_b = (vertex_d.x, vertex_d.y)
        coord_c = (vertex_b.x, vertex_b.y)
        coord_d = (vector_ref.x, vector_ref.y)
    else:
        a1, a2, a3 = set_mode(plane)
        coord_a = (vertex_c[a1], vertex_c[a2])
        coord_b = (vertex_d[a1], vertex_d[a2])
        coord_c = (vertex_a[a1], vertex_a[a2])
        coord_d = (vertex_b[a1], vertex_b[a2])
    v_stack = np.vstack([coord_a, coord_b, coord_c, coord_d])
    h_stack = np.hstack((v_stack, np.ones((4, 1))))
    line_a = np.cross(h_stack[0], h_stack[1])
    line_b = np.cross(h_stack[2], h_stack[3])
    x_loc, y_loc, z_loc = np.cross(line_a, line_b)
    if z_loc == 0:
        return Vector((0, 0, 0)), False
    new_x_loc = x_loc / z_loc
    new_z_loc = y_loc / z_loc
    if plane == "LO":
        new_y_loc = 0
    else:
        new_y_loc = vertex_a[a3]
    # Order Vector Delta
    if plane == "XZ":
        vector_delta = Vector((new_x_loc, new_y_loc, new_z_loc))
    elif plane == "XY":
        vector_delta = Vector((new_x_loc, new_z_loc, new_y_loc))
    elif plane == "YZ":
        vector_delta = Vector((new_y_loc, new_x_loc, new_z_loc))
    else:
        # Must be Local View Plane
        vector_delta = view_coords(new_x_loc, new_z_loc, new_y_loc) + vertex_a
    return vector_delta, True


def get_percent(obj, flip_percent, per_v, data, scene):
    """Calculates a Percentage Distance between 2 Vectors.

    Note:
        Calculates a point that lies a set percentage between two given points
        using standard Numpy Routines.

        Works for either 2 vertices for an object in Edit mode
        or 2 selected objects in Object mode.

    Args:
        obj: The Object under consideration
        flip_percent: Setting this to True measures the percentage starting from the second vector
        per_v: Percentage Input Value
        data: pg.flip, pg.percent scene variables & Operational Mode
        scene: Context Scene

    Returns:
        World Vector.
    """

    pg = scene.pdt_pg

    if obj.mode == "EDIT":
        bm = bmesh.from_edit_mesh(obj.data)
        verts = [v for v in bm.verts if v.select]
        if len(verts) == 2:
            vector_a = verts[0].co
            vector_b = verts[1].co
            if vector_a is None:
                pg.error = PDT_ERR_VERT_MODE
                bpy.context.window_manager.popup_menu(oops, title="Error", icon="ERROR")
                return None
        else:
            pg.error = PDT_ERR_SEL_2_V_1_E + str(len(verts)) + " Vertices"
            bpy.context.window_manager.popup_menu(oops, title="Error", icon="ERROR")
            return None
        coord_a = np.array([vector_a.x, vector_a.y, vector_a.z])
        coord_b = np.array([vector_b.x, vector_b.y, vector_b.z])
    if obj.mode == "OBJECT":
        objs = bpy.context.view_layer.objects.selected
        if len(objs) != 2:
            pg.error = PDT_ERR_SEL_2_OBJS + str(len(objs)) + ")"
            bpy.context.window_manager.popup_menu(oops, title="Error", icon="ERROR")
            return None
        coord_a = np.array(
            [
                objs[-1].matrix_world.decompose()[0].x,
                objs[-1].matrix_world.decompose()[0].y,
                objs[-1].matrix_world.decompose()[0].z,
            ]
        )
        coord_b = np.array(
            [
                objs[-2].matrix_world.decompose()[0].x,
                objs[-2].matrix_world.decompose()[0].y,
                objs[-2].matrix_world.decompose()[0].z,
            ]
        )
    coord_c = coord_b - coord_a
    coord_d = np.array([0, 0, 0])
    _per_v = per_v
    if (flip_percent and data != "MV") or data == "MV":
        _per_v = 100 - per_v
    coord_out = (coord_d+coord_c) * (_per_v / 100) + coord_a
    return Vector((coord_out[0], coord_out[1], coord_out[2]))


def obj_check(obj, scene, operation):
    """Check Object & Selection Validity.

    Args:
        obj: Active Object
        scene: Active Scene
        operation: The Operation e.g. Create New Vertex

    Returns:
        Object Bmesh
        Validity Boolean.
    """

    pg = scene.pdt_pg
    _operation = operation.upper()

    if obj is None:
        pg.error = PDT_ERR_NO_ACT_OBJ
        bpy.context.window_manager.popup_menu(oops, title="Error", icon="ERROR")
        return None, False
    if obj.mode == "EDIT":
        bm = bmesh.from_edit_mesh(obj.data)
        if _operation == "S":
            if len(bm.edges) < 1:
                pg.error = f"{PDT_ERR_SEL_1_EDGEM} {len(bm.edges)})"
                bpy.context.window_manager.popup_menu(oops, title="Error", icon="ERROR")
                return None, False
            return bm, True
        if len(bm.select_history) >= 1:
            vector_a = None
            if _operation not in {"D", "E", "F", "G", "N", "S"}:
                vector_a = check_selection(1, bm, obj)
            else:
                verts = [v for v in bm.verts if v.select]
                if len(verts) > 0:
                    vector_a = verts[0]
            if vector_a is None:
                pg.error = PDT_ERR_VERT_MODE
                bpy.context.window_manager.popup_menu(oops, title="Error", icon="ERROR")
                return None, False
        return bm, True
    return None, True


def dis_ang(values, flip_angle, plane, scene):
    """Set Working Axes when using Direction command.

    Args:
        values: Input Arguments
        flip_angle: Whether to flip the angle
        plane: Working Plane
        scene: Current Scene

    Returns:
        Directional Offset as a Vector.
    """

    pg = scene.pdt_pg
    dis_v = float(values[0])
    ang_v = float(values[1])
    if flip_angle:
        if ang_v > 0:
            ang_v = ang_v - 180
        else:
            ang_v = ang_v + 180
        pg.angle = ang_v
    if plane == "LO":
        vector_delta = view_dir(dis_v, ang_v)
    else:
        a1, a2, _ = set_mode(plane)
        vector_delta = Vector((0, 0, 0))
        # fmt: off
        vector_delta[a1] = vector_delta[a1] + (dis_v * cos(ang_v * pi/180))
        vector_delta[a2] = vector_delta[a2] + (dis_v * sin(ang_v * pi/180))
        # fmt: on
    return vector_delta


# Shader for displaying the Pivot Point as Graphics.
#
SHADER = gpu.shader.from_builtin("UNIFORM_COLOR") if not bpy.app.background else None


def draw_3d(coords, gtype, rgba, context):
    """Draw Pivot Point Graphics.

    Note:
        Draws either Lines Points, or Tris using defined shader

    Args:
        coords: Input Coordinates List
        gtype: Graphic Type
        rgba: Colour in RGBA format
        context: Blender bpy.context instance.

    Returns:
        Nothing.
    """

    batch = batch_for_shader(SHADER, gtype, {"pos": coords})

    try:
        if coords is not None:
            gpu.state.blend_set('ALPHA')
            SHADER.bind()
            SHADER.uniform_float("color", rgba)
            batch.draw(SHADER)
    except:
        raise PDT_ShaderError


def draw_callback_3d(self, context):
    """Create Coordinate List for Pivot Point Graphic.

    Note:
        Creates coordinates for Pivot Point Graphic consisting of 6 Tris
        and one Point colour coded Red; X axis, Green; Y axis, Blue; Z axis
        and a yellow point based upon screen scale

    Args:
        context: Blender bpy.context instance.

    Returns:
        Nothing.
    """

    scene = context.scene
    pg = scene.pdt_pg
    region_width = context.region.width
    x_loc = pg.pivot_loc.x
    y_loc = pg.pivot_loc.y
    z_loc = pg.pivot_loc.z
    # Scale it from view
    areas = [a for a in context.screen.areas if a.type == "VIEW_3D"]
    if len(areas) > 0:
        scale_factor = abs(areas[0].spaces.active.region_3d.window_matrix.decompose()[2][1])
    # Check for orhtographic view and resize
    #if areas[0].spaces.active.region_3d.is_orthographic_side_view:
    #    dim_a = region_width / sf / 60000 * pg.pivot_size
    #else:
    #    dim_a = region_width / sf / 5000 * pg.pivot_size
    dim_a = region_width / scale_factor / 50000 * pg.pivot_size
    dim_b = dim_a * 0.65
    dim_c = dim_a * 0.05 + (pg.pivot_width * dim_a * 0.02)
    dim_o = dim_c / 3

    # fmt: off
    # X Axis
    coords = [
        (x_loc, y_loc, z_loc),
        (x_loc+dim_b, y_loc-dim_o, z_loc),
        (x_loc+dim_b, y_loc+dim_o, z_loc),
        (x_loc+dim_a, y_loc, z_loc),
        (x_loc+dim_b, y_loc+dim_c, z_loc),
        (x_loc+dim_b, y_loc-dim_c, z_loc),
    ]
    # fmt: on
    colour = (1.0, 0.0, 0.0, pg.pivot_alpha)
    draw_3d(coords, "TRIS", colour, context)
    coords = [(x_loc, y_loc, z_loc), (x_loc+dim_a, y_loc, z_loc)]
    draw_3d(coords, "LINES", colour, context)
    # fmt: off
    # Y Axis
    coords = [
        (x_loc, y_loc, z_loc),
        (x_loc-dim_o, y_loc+dim_b, z_loc),
        (x_loc+dim_o, y_loc+dim_b, z_loc),
        (x_loc, y_loc+dim_a, z_loc),
        (x_loc+dim_c, y_loc+dim_b, z_loc),
        (x_loc-dim_c, y_loc+dim_b, z_loc),
    ]
    # fmt: on
    colour = (0.0, 1.0, 0.0, pg.pivot_alpha)
    draw_3d(coords, "TRIS", colour, context)
    coords = [(x_loc, y_loc, z_loc), (x_loc, y_loc + dim_a, z_loc)]
    draw_3d(coords, "LINES", colour, context)
    # fmt: off
    # Z Axis
    coords = [
        (x_loc, y_loc, z_loc),
        (x_loc-dim_o, y_loc, z_loc+dim_b),
        (x_loc+dim_o, y_loc, z_loc+dim_b),
        (x_loc, y_loc, z_loc+dim_a),
        (x_loc+dim_c, y_loc, z_loc+dim_b),
        (x_loc-dim_c, y_loc, z_loc+dim_b),
    ]
    # fmt: on
    colour = (0.2, 0.5, 1.0, pg.pivot_alpha)
    draw_3d(coords, "TRIS", colour, context)
    coords = [(x_loc, y_loc, z_loc), (x_loc, y_loc, z_loc + dim_a)]
    draw_3d(coords, "LINES", colour, context)
    # Centre
    coords = [(x_loc, y_loc, z_loc)]
    colour = (1.0, 1.0, 0.0, pg.pivot_alpha)
    draw_3d(coords, "POINTS", colour, context)


def scale_set(self, context):
    """Sets Scale by dividing Pivot Distance by System Distance.

    Note:
        Sets Pivot Point Scale Factors by Measurement
        Uses pg.pivotdis & pg.distance scene variables

    Args:
        context: Blender bpy.context instance.

    Returns:
        Status Set.
    """

    scene = context.scene
    pg = scene.pdt_pg
    sys_distance = pg.distance
    scale_distance = pg.pivot_dis
    if scale_distance > 0:
        scale_factor = scale_distance / sys_distance
        pg.pivot_scale = Vector((scale_factor, scale_factor, scale_factor))



# SPDX-FileCopyrightText: 2019-2023 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later

import bpy
from bpy.types import Operator, Menu
from bl_operators.presets import AddPresetBase
from bl_ui.utils import PresetPanel
import os
from math import degrees

from .sun_calc import format_lat_long, format_time, format_hms, sun


# -------------------------------------------------------------------
# Choice list of places, month and day at 12:00 noon
# -------------------------------------------------------------------


class SUNPOS_PT_Presets(PresetPanel, bpy.types.Panel):
    bl_label = "Sun Position Presets"
    preset_subdir = "operator/sun_position"
    preset_operator = "script.execute_preset"
    preset_add_operator = "world.sunpos_add_preset"


class SUNPOS_OT_AddPreset(AddPresetBase, Operator):
    '''Add Sun Position preset'''
    bl_idname = "world.sunpos_add_preset"
    bl_label = "Add Sun Position preset"
    preset_menu = "SUNPOS_PT_Presets"

    # variable used for all preset values
    preset_defines = [
        "sun_props = bpy.context.scene.sun_pos_properties"
    ]

    # properties to store in the preset
    preset_values = [
        "sun_props.day",
        "sun_props.month",
        "sun_props.time",
        "sun_props.year",
        "sun_props.UTC_zone",
        "sun_props.use_daylight_savings",
        "sun_props.latitude",
        "sun_props.longitude",
    ]

    # where to store the preset
    preset_subdir = "operator/sun_position"


# -------------------------------------------------------------------
#
#   Draw the Sun Panel, sliders, et. al.
#
# -------------------------------------------------------------------

class SUNPOS_PT_Panel(bpy.types.Panel):
    bl_space_type = "PROPERTIES"
    bl_region_type = "WINDOW"
    bl_context = "world"
    bl_label = "Sun Position"
    bl_options = {'DEFAULT_CLOSED'}

    def draw_header_preset(self, _context):
        SUNPOS_PT_Presets.draw_panel_header(self.layout)

    def draw(self, context):
        sun_props = context.scene.sun_pos_properties
        layout = self.layout
        layout.use_property_split = True
        layout.use_property_decorate = False

        layout.prop(sun_props, "usage_mode", expand=True)
        layout.separator()

        if sun_props.usage_mode == "HDR":
            self.draw_environment_mode_panel(context)
        else:
            self.draw_normal_mode_panel(context)

    def draw_environment_mode_panel(self, context):
        sun_props = context.scene.sun_pos_properties
        layout = self.layout

        col = layout.column(align=True)
        col.prop_search(sun_props, "sun_object",
                        context.view_layer, "objects")
        if context.scene.world is not None:
            if context.scene.world.node_tree is not None:
                col.prop_search(sun_props, "hdr_texture",
                                context.scene.world.node_tree, "nodes")
            else:
                col.label(text="Please activate Use Nodes in the World panel.",
                          icon="ERROR")
        else:
            col.label(text="Please select World in the World panel.",
                      icon="ERROR")

        layout.use_property_decorate = True

        col = layout.column(align=True)
        col.prop(sun_props, "bind_to_sun", text="Bind Texture to Sun")
        col.prop(sun_props, "hdr_azimuth")
        row = col.row(align=True)
        row.active = not sun_props.bind_to_sun
        row.prop(sun_props, "hdr_elevation")
        col.prop(sun_props, "sun_distance")
        col.separator()

        col = layout.column(align=True)
        row = col.row(align=True)
        row.enabled = not sun_props.bind_to_sun
        row.operator("world.sunpos_show_hdr", icon='LIGHT_SUN')

    def draw_normal_mode_panel(self, context):
        sun_props = context.scene.sun_pos_properties
        addon_prefs = context.preferences.addons[__package__].preferences
        layout = self.layout

        col = layout.column(align=True)
        col.prop(sun_props, "sun_object")
        col.separator()

        col.prop(sun_props, "object_collection")
        if sun_props.object_collection:
            col.prop(sun_props, "object_collection_type")
            if sun_props.object_collection_type == 'DIURNAL':
                col.prop(sun_props, "time_spread")
        col.separator()

        if context.scene.world is not None:
            if context.scene.world.node_tree is not None:
                col.prop_search(sun_props, "sky_texture",
                                context.scene.world.node_tree, "nodes")
            else:
                col.label(text="Please activate Use Nodes in the World panel.",
                          icon="ERROR")
        else:
            col.label(text="Please select World in the World panel.",
                      icon="ERROR")

        if addon_prefs.show_overlays:
            col = layout.column(align=True, heading="Show")
            col.prop(sun_props, "show_north", text="North")
            col.prop(sun_props, "show_analemmas", text="Analemmas")
            col.prop(sun_props, "show_surface", text="Surface")

        if addon_prefs.show_refraction:
            col = layout.column(align=True, heading="Use")
            col.prop(sun_props, "use_refraction", text="Refraction")


class SUNPOS_PT_Location(bpy.types.Panel):
    bl_space_type = "PROPERTIES"
    bl_region_type = "WINDOW"
    bl_context = "world"
    bl_label = "Location"
    bl_parent_id = "SUNPOS_PT_Panel"

    @classmethod
    def poll(self, context):
        sun_props = context.scene.sun_pos_properties
        return sun_props.usage_mode != "HDR"

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        sun_props = context.scene.sun_pos_properties
        addon_prefs = context.preferences.addons[__package__].preferences

        col = layout.column(align=True)
        col.prop(sun_props, "coordinates", icon='URL')
        col.prop(sun_props, "latitude")
        col.prop(sun_props, "longitude")

        col.separator()

        col = layout.column(align=True)
        col.prop(sun_props, "north_offset", text="North Offset")

        if addon_prefs.show_az_el:
            col = layout.column(align=True)
            col.prop(sun_props, "sun_elevation", text="Elevation")
            col.prop(sun_props, "sun_azimuth", text="Azimuth")
            col.separator()

        col = layout.column()
        col.prop(sun_props, "sun_distance")
        col.separator()


class SUNPOS_PT_Time(bpy.types.Panel):
    bl_space_type = "PROPERTIES"
    bl_region_type = "WINDOW"
    bl_context = "world"
    bl_label = "Time"
    bl_parent_id = "SUNPOS_PT_Panel"

    @classmethod
    def poll(self, context):
        sun_props = context.scene.sun_pos_properties
        return sun_props.usage_mode != "HDR"

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        sun_props = context.scene.sun_pos_properties
        addon_prefs = context.preferences.addons[__package__].preferences

        col = layout.column(align=True)
        col.prop(sun_props, "use_day_of_year")
        if sun_props.use_day_of_year:
            col.prop(sun_props, "day_of_year")
        else:
            col.prop(sun_props, "day")
            col.prop(sun_props, "month")
        col.prop(sun_props, "year")
        col.separator()

        col = layout.column(align=True)
        col.prop(sun_props, "time", text="Time", text_ctxt="Hour")
        col.prop(sun_props, "UTC_zone")
        col.prop(sun_props, "use_daylight_savings")
        col.separator()

        local_time = format_time(sun_props.time,
                                 sun_props.use_daylight_savings)
        utc_time = format_time(sun_props.time,
                               sun_props.use_daylight_savings,
                               sun_props.UTC_zone)

        col = layout.column(align=True)
        col.alignment = 'CENTER'

        split = col.split(factor=0.5, align=True)
        sub = split.column(align=True)
        sub.alignment = 'RIGHT'
        sub.label(text="Time Local:")
        sub.label(text="UTC:")

        sub = split.column(align=True)
        sub.label(text=local_time)
        sub.label(text=utc_time)
        col.separator()

        if addon_prefs.show_rise_set:
            sunrise = format_hms(sun.sunrise)
            sunset = format_hms(sun.sunset)

            col = layout.column(align=True)
            col.alignment = 'CENTER'

            split = col.split(factor=0.5, align=True)
            sub = split.column(align=True)
            sub.alignment = 'RIGHT'
            sub.label(text="Sunrise:")
            sub.label(text="Sunset:")

            sub = split.column(align=True)
            sub.label(text=sunrise)
            sub.label(text=sunset)

        col.separator()
